#!/usr/bin/python3

#	ax.alt : Lib that helps to work with ALT Linux repos
#	Copyright (C) 2019-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import os
import re
from collections    import defaultdict
from ax.sefunctions import CheckFile

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get name, version and release of all packages of a given branch

def PackageNames(branch, space=None):

	if not space:
		space = '/space'

	src_list = os.path.join(space, 'ALT', branch, 'files/list/src.list')

	if not CheckFile(src_list, os.R_OK):
		return {}

	packages = defaultdict(dict)

	with open(src_list, 'r') as f:
		for line in f.readlines():
			search = re.search(r'^\s?'
				'(?P<name>[^\s]+)\s+'
				'(?P<epoch_ver_rel>[^\s]+)\s+'
				'(?P<src>[^\s]+)\s?$',
				line)
			# Cutting epoch and splitting version and release
			version, release = VerRel(search.group('epoch_ver_rel'))
			packages[search.group('name')] = {
				'version': version,
				'release': release
				}

	return packages

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Split full name of a package to [name, version, release] if two_parts flag
# is not given or to [name, version] otherwise

def SplitPackageName(package, two_parts=False):

	package = re.sub(r'\.src$', '', re.sub(r'\.rpm$', '', package))
	delim_index  = -1
	delim_index2 = -1
	n_delims = 1 if two_parts else 2

	for i in range(n_delims):
		delim_index = package.rfind('-', 0, delim_index)
		if delim_index > 0:
			if delim_index2 < 0:
				delim_index2 = delim_index
		else:
			delim_index = None
			break

	splitted_package_name = [''] * (n_delims + 1)

	if not delim_index:
		if delim_index2 < 0:
			splitted_package_name[0] = package
		else:
			splitted_package_name[0] = package[:delim_index2]
			splitted_package_name[1] = package[(delim_index2 + 1):]
	else:
		splitted_package_name[0] = package[:delim_index]
		if two_parts or delim_index2 < 0:
			splitted_package_name[1] = package[(delim_index + 1):]
		else:
			splitted_package_name[1] = package[(delim_index + 1):delim_index2]
			splitted_package_name[2] = package[(delim_index2 + 1):]

	return splitted_package_name

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get version and release (no epoch) of a given version

def VerRel(ver_rel):

	# Stripping an epoch
	ver_rel = StripEpoch(ver_rel)

	# Splitting a version and a release
	match = re.match(r'(?P<ver>[^-]+)-(?P<rel>[^-]+)', ver_rel)

	if not match:
		return ('', '')

	return (match.group('ver'), match.group('rel'))

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Strip epoch from the left side

def StripEpoch(epoch_ver_rel):

	epoch = re.sub(r'[^:]+$', '', epoch_ver_rel)

	if epoch:
		epoch = epoch[:-1]
		return epoch_ver_rel[len(epoch) + 1:]

	return epoch_ver_rel
