#!/usr/bin/python3

#	ax.printer : Lib for developing clean and pretty CLI
#	Copyright (C) 2019-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import sys
import os
from collections import defaultdict

DEFAULT_TAB_SIZE = 4

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# The Printer class presents special ways of handling the output

class Printer():

	def __init__(self, silent=None, alerts_only=None, plain=None,
			block_printing=None):

		self.preface = ''    # Preface that will be displayed before every line
		self.tie = '.'       # Symbol used to tie l and r parts of the status line
		self.indent = []     # Indentation stack
		self.operation = ''  # Left part of the status line
		self.operation0 = '' # Prev operation
		self.pages = defaultdict(str) # Marked multiline text
		self.page_mark = ''  # Current page mark used to write into 'page' dict
		self.memorizing = defaultdict(bool) # Writing into mem instead of stdout
		self.alert = False   # Current message is err or warn message
		self.numb  = False   # Don't do anything
		self.extra = []      # Extra info displayed under the main status line

		# Enabling/disabling the output
		self.silent = silent if silent != None else False
		# Displaying only err and warn messages
		self.alerts_only = alerts_only if alerts_only != None else False
		# Swithcing plain output on/off
		self.plain = plain if plain != None else False
		# Displaying beginning of the line right away
		self.block_printing = block_printing if (block_printing != None and
			not self.alerts_only) else False

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	# Get terminal width
	def __GetTerminalWidth(self):
		# Pretending that terminal width is 80 symbols if plain mode is activated
		#	or if stdout is not connected to a terminal
		if self.plain or not os.isatty(sys.stdout.fileno()):
			return 80
		return int(os.popen('stty size', 'r').read().split()[-1])

	# Memorize or print
	def __Print(self, string):
		if self.silent or (self.alerts_only and not self.alert):
			return
		if self.memorizing[self.page_mark]:
			self.pages[self.page_mark] += string
		else:
			print(string, end='')
			sys.stdout.flush()

	# Print joined indentation units
	def __PrintIndent(self):
		for i in self.indent:
			self.__Print(i);

	# Print given string at the right part of the terminal and
	#	tie it with the left part by the row of 'tie' symbols
	def __CompleteTheLine(self, status):
		if self.silent or (self.alerts_only and not self.alert):
			return
		tie = ''
		if self.operation:
			if self.plain:
				tie = ' '
			else:
				term_width = self.__GetTerminalWidth()
				# Number of colums between printed text and new text
				tie_len = term_width - len(self.operation) - len(status) - \
					sum(len(indent) for indent in self.indent)
				while tie_len < 0:
					tie_len += term_width
				tie = self.tie * tie_len
		else:
			# Indent was not printed because a current value of the 'operation'
			# variable is an empty string
			self.__PrintIndent()
		self.__Print(tie + status + '\n')
		# Displaying extra info
		for ex in self.extra:
			if ex:
				self.__Print(f'\t- {ex}\n')

	# Forming special message
	def __SpecMessage(self, preface, msg=''):
		return f'[{preface}{": " + msg if msg else ""}]'

	# Clearing data fields
	def __Clear(self):
		self.operation0 = self.operation
		self.operation = ''
		self.extra = []
		self.alert = False

	# Raise one level of indentation (and optionally change size of all indents)
	def IndentPush(self, size=DEFAULT_TAB_SIZE):
		self.indent.append(' ' * (size if size > 0 else DEFAULT_TAB_SIZE))

	# Drop one level of indentation
	def IndentPop(self):
		if self.indent:
			self.indent.pop()

	# Clear the stack of indentation strings
	def IndentClear(self):
		self.indent.clear()

	# Get current left part of the status line
	def Operation(self):
		return self.operation

	# Get vector of extra lines
	def Extra(self):
		return self.extra

	# Write 'silent' (do no print) flag
	def Silent(self, flag):
		self.silent = flag

	# Get 'silent' (do no print) flag
	def IsSilent(self):
		return self.silent

	# Write 'numb' (don't do anything) flag
	def Numb(self, flag):
		self.numb = flag

	# Define tie symbol
	def TieSymbol(self, s):
		self.tie = s

	# Write the preface that will be displayed before every line
	def SetPreface(self, preface=''):
		self.preface = f'*{preface}* ' if preface else ''

	# Drop preface
	def DropPreface(self):
		self.preface = ''

	# Start a new status line
	def LineBegin(self, msg):
		if self.numb:
			return
		if self.block_printing:
			self.PageBegin()
		if self.operation:
			self.LineBreak()
		if msg:
			self.__PrintIndent()
			self.operation = (self.preface + msg) if self.preface else msg
			self.__Print(self.operation)

	# Concatenate with a current status line
	def LineCat(self, msg):
		if self.numb:
			return
		if not self.operation:
			self.LineBegin(msg)
			return
		self.operation += msg
		self.__Print(msg)

	# Complete current status line
	def LineEnd(self, msg):
		if self.numb:
			return
		self.__CompleteTheLine(msg)
		if self.block_printing:
			self.__Print(self.PageEnd())
		self.__Clear()

	# Print success message
	def Success(self, msg=''):
		self.LineEnd(self.__SpecMessage('DONE', msg))

	# Printing warning message
	def Warn(self, msg=''):
		self.alert = True
		self.LineEnd(self.__SpecMessage('WARNING', msg))

	# Printing error message
	def Err(self, msg=''):
		self.alert = True
		self.LineEnd(self.__SpecMessage('ERROR', msg))

	# Printing list of lines
	def Multiline(self, lines):
		for line in lines:
			self.LineEnd(line)

	# Add extra line, displayed under the main status line
	def LineAddExtra(self, msg):
		if self.numb:
			return
		self.extra.append(msg)

	# Add extra warning line, displayed under the main status line
	def LineAddWarn(self, msg):
		self.LineAddExtra(self.__SpecMessage('WARNING', msg))

	# Add extra error line, displayed under the main status line
	def LineAddErr(self, msg):
		self.LineAddExtra(self.__SpecMessage('ERROR', msg))

	# Break current status line
	def LineBreak(self):
		if self.numb:
			return
		self.__Clear()
		self.__Print((self.PageEnd() if self.block_printing else '') + '\n')

	# Begin the new line with the same message on the left side
	def LineRestart(self):
		if self.operation:
			self.LineBreak()
		self.LineBegin(self.operation0)

	# Start to memorize
	def PageBegin(self, mark=None):
		if mark != None:
			self.page_mark = mark
		self.memorizing[self.page_mark] = True
		self.pages[self.page_mark] = ''

	# Start to memorize
	def PageEnd(self, mark=None):
		if mark == None:
			mark = self.page_mark
		self.memorizing[mark] = False
		res = self.pages[mark]
		self.pages[mark] = ''
		return res

	# Print divider on entire terminal line
	def Div(self, s='='):
		if self.numb:
			return
		if self.silent or (self.alerts_only and not self.alert):
			return
		term_width = self.__GetTerminalWidth()
		divider = s * term_width
		self.__Print(divider + '\n')

	# Print text betwern two dividers
	def Distinguish(self, text):
		if self.numb:
			return
		self.Div()
		self.__Print(text + '\n')
		self.Div()

	# Print DONE or ERROR status
	def BiStatus(self, condition, msg=''):
		if condition:
			self.Success(msg)
			return True
		else:
			self.Err(msg)
			return False
