#!/usr/bin/python3

#	ax.ver : Lib that helps to work with software versions
#	Copyright (C) 2019-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from datetime         import datetime
from ax.purefunctions import ReplaceNotAlfa

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

CUR_Y = datetime.today().year
CUR_DC = (CUR_Y - 2000) // 10
PREV_DC = CUR_DC - 1
Y_OF_A_CUR_DC = CUR_Y - 2000 - CUR_DC * 10
Y_REGEX = fr'(20({CUR_DC}[0-{Y_OF_A_CUR_DC}]|[0-{PREV_DC}][0-9])|199[1-9])'
M_REGEX = r'(0[1-9]|1[0-2])'
D_REGEX = r'(0[1-9]|[1-2][0-9]|3[0-1])'
DATE_VER_REGEX = re.compile(fr'^{Y_REGEX}')
DATE_REGEX = re.compile(
	r'(?P<pre>.*)'
	r'(?<![0-9])'
	fr'(?P<year>{Y_REGEX})\.?'
	fr'(?P<month>{M_REGEX})\.?'
	fr'(?P<day>{D_REGEX})?'
	r'(?P<post>.*)'
	)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Check whether a given version contains a date at the beginning

def IsDateVer(ver_str):

	if not ver_str:
		return False

	# If value of the argument is a range of versions
	if ver_str[0] in '[(':
		# It is a date-version if at least one side contains a date
		return any(IsDateVer(el) for el in ver_str.strip('[()]').split(';'))

	if len(ver_str) < 4:
		return False

	return re.search(DATE_VER_REGEX, ver_str) != None

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Convert date-part of a version string into a "yyyy.mm.dd" format and separate
# it with dots from everything that placed pre and post

def ConvDate(ver_str):

	match = re.match(DATE_REGEX, ver_str)
	if not match:
		return ver_str

	pre  = match.group('pre')
	y    = match.group('year')
	m    = match.group('month')
	d    = match.group('day')
	post = match.group('post')

	return (f'{pre}.' if pre and pre[-1] != "." else pre) + \
		f'{y}.{m}.{d if d else "01"}' + \
		(f'.{post}' if post and post[0] != "." else post)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Convert version represented by string into a list of subversions

def SVerToLVer(ver_str):

	if not ver_str:
		return []

	# Replacing delimiters and converting a date (if it's there)
	converted_ver_str = ConvDate(
		ReplaceNotAlfa(ver_str, '.', allow_in_versions=False, exceptions={'*'})
		)

	return converted_ver_str.split('.')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Compare two versions represented by strings

def CompareSVersions(ver1, ver2):

	return CompareLVersions(SVerToLVer(ver1), SVerToLVer(ver2))

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Compare two versions represented by lists, returned values are:
# res - index of a critical part of compared versions (starting from 1),
#	negative if ver1 < ver2, positive if ver1 > ver2, 0 if ver1 == ver2;
# symbolic - 0 or 3 if both versions are not symbolic or are symbolic resp.,
#	1 or 2 if a first ver is sybolic and a second one is not and vice versa

# Get numeric part from the left side of a given string
def __GetNumeric(s):
	search = re.search('^[0-9]+', s)
	return int(search.group()) if search else -1

def CompareLVersions(ver1, ver2):

	res = 0
	symbolic = 0
	l1 = len(ver1)
	l2 = len(ver2)
	lmin = min(l1, l2)
	same_len = l1 == l2

	for i in range(0, lmin):
		# Comparison of complying sections
		ver1_isdigit = ver1[i].isdigit()
		ver2_isdigit = ver2[i].isdigit()
		if ver1_isdigit and ver2_isdigit:
			ver1_i = int(ver1[i])
			ver2_i = int(ver2[i])
		else:
			if any(ver == '*' for ver in (ver1[i], ver2[i])):
				break
			skip = False
			if same_len and i == (lmin - 1):
				ver1_i_numpart = __GetNumeric(ver1[i]) if not ver1_isdigit \
					else int(ver1[i])
				ver2_i_numpart = __GetNumeric(ver2[i]) if not ver2_isdigit \
					else int(ver2[i])
				if ver1_i_numpart > 0 and ver2_i_numpart > 0 and \
						ver1_i_numpart != ver2_i_numpart:
					skip = True
					ver1_i = ver1_i_numpart
					ver2_i = ver2_i_numpart
			if not skip:
				ver1_i = ver1[i]
				ver2_i = ver2[i]
				if not ver1_isdigit:
					symbolic |= 0b01
				if not ver2_isdigit:
					symbolic |= 0b10
		if ver1_i > ver2_i:
			res = i + 1
			break
		if ver1_i < ver2_i:
			res = -(i + 1)
			break

	return (res, symbolic)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Compare a single ver and a range of versions both represented by strings,
# range is '[<optional_str>;<optional_str]', '(<opt_str>;<opt_str>]' etc,
# return values are:
# comp - index of a critical part of compared versions (starting from 1),
#	negative if ver is lesser than both limits of ver_range, positive if ver is
#	greater than both limits of ver_range, 0 if ver is included in ver_range;
# symbolic_left - 0 or 3 if both a single version and a left part of a range
#	are not symbolic or both are symbolic resp., 1 or 2 if a single version is
#	sybolic and a left part of a range is not and vice versa;
# symbolic_left - 0 or 3 if ... a right part of a range ...

def CompareSVerWithSRange(ver, ver_range):

	if not ver or not ver_range or len(ver_range) < 3:
		return [None] * 3

	ver_range_pair = ver_range.strip('[()]').split(';')
	if len(ver_range_pair) != 2:
		return [None] * 3

	range_left  = ver_range_pair[0]
	range_right = ver_range_pair[1]

	comp_left, symbolic_left = (1, 0) \
		if (not range_left or range_left == '*') \
		else CompareSVersions(ver, range_left)

	comp_right, symbolic_right = (-1, 0) \
		if (not range_right or range_right == '*') \
		else CompareSVersions(ver, range_right)

	start_incl = (ver_range[0]  == '[')
	end_incl   = (ver_range[-1] == ']')

	if comp_left < 0:
		comp = comp_left
	elif comp_left == 0 and not start_incl:
		comp = -1
	elif comp_right > 0:
		comp = comp_right
	elif comp_right == 0 and not end_incl:
		comp = 1
	else:
		comp = 0

	return comp, symbolic_left, symbolic_right

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Define type of a version record

def GetSingleVerAndRLeftRRight(single_ver_or_range):

	single_ver  = None
	range_left  = None
	range_right = None

	if single_ver_or_range:
		if single_ver_or_range[0] in ('[', '(') and \
				single_ver_or_range[-1] in (']', ')'):
			range_pair = single_ver_or_range.strip('[()]').split(';')
			if len(range_pair) == 2:
				range_left  = range_pair[0]
				range_right = range_pair[1]
		elif single_ver_or_range[0] not in ('[', '(') and \
				single_ver_or_range[-1] not in (']', ')'):
			single_ver = single_ver_or_range

	return (single_ver, range_left, range_right)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Merge single versions and ranges of versions from a given list

def MergeVersions(sversions_and_sranges):

	if not sversions_and_sranges:
		return sversions_and_sranges

	merged = [sversions_and_sranges[0]]

	for single_ver_or_range in sversions_and_sranges[1:]:
		single_ver, range_left, range_right = \
			GetSingleVerAndRLeftRRight(single_ver_or_range)
		current_merge = []
		# If flag will remain false until the end of the loop,
		#	then currently processed version record hasn't already been merged
		#	and should be appended
		flag = False
		for m_single_ver_or_range in merged:
			m_single_ver, m_range_left, m_range_right = \
				GetSingleVerAndRLeftRRight(m_single_ver_or_range)
			# Now types of version records are determined,
			#	and the merging action can be chosen
			if single_ver != None:
				if m_single_ver != None:
					if CompareSVersions(single_ver, m_single_ver)[0] != 0:
						current_merge.append(m_single_ver)
				elif m_range_left != None or m_range_right != None:
					comp = CompareSVerWithSRange(single_ver, m_single_ver_or_range)[0]
					if comp == None:
						continue
					if comp == 0:
						current_merge.append(m_single_ver_or_range)
						flag = True
					elif comp < 0:
						if CompareSVersions(single_ver, m_range_left)[0] == 0:
							new_range = '[' + m_single_ver_or_range[1:]
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver_or_range)
					else:
						if CompareSVersions(single_ver, m_range_right)[0] == 0:
							new_range = m_single_ver_or_range[:-1] + ']'
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver_or_range)
			elif range_left != None or range_right != None:
				if m_single_ver != None:
					comp = CompareSVerWithSRange(m_single_ver, single_ver_or_range)[0]
					if comp == None:
						continue
					if comp == 0:
						current_merge.append(single_ver_or_range)
						flag = True
					elif comp < 0:
						if CompareSVersions(m_single_ver, range_left)[0] == 0:
							new_range = '[' + single_ver_or_range[1:]
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver)
					else:
						if CompareSVersions(m_single_ver, range_right)[0] == 0:
							new_range = single_ver_or_range[:-1] + ']'
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver)
				elif m_range_left != None or m_range_right != None:
					comp_l = -1
					comp_r = 1
					if range_left:
						comp_l = CompareSVerWithSRange(range_left,  m_single_ver_or_range)[0]
					if range_right:
						comp_r = CompareSVerWithSRange(range_right, m_single_ver_or_range)[0]
					if comp_l == None or comp_r == None:
						continue
					if comp_l == 0 and comp_r == 0:
						current_merge.append(m_single_ver_or_range)
						flag = True
					if comp_l < 0 and comp_r > 0:
						continue
					elif comp_r < 0:
						if CompareSVersions(range_right, m_range_left)[0] == 0:
							new_range = single_ver_or_range[0] + \
								range_left + ';' + m_range_right + \
								m_single_ver_or_range[-1]
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver_or_range)
					elif comp_l > 0:
						if CompareSVersions(m_range_right, range_left)[0] == 0:
							new_range = m_single_ver_or_range[0] + \
								m_range_left + ';' + range_right + \
								single_ver_or_range[-1]
							current_merge.append(new_range)
							flag = True
						else:
							current_merge.append(m_single_ver_or_range)
					elif comp_r == 0 and comp_l < 0:
						new_range = single_ver_or_range[0] + range_left + \
							';' + m_range_right + m_single_ver_or_range[-1]
						current_merge.append(new_range)
						flag = True
					elif comp_l == 0 and comp_r > 0:
						new_range = m_single_ver_or_range[0] + m_range_left + \
							';' + range_right + single_ver_or_range[-1]
						current_merge.append(new_range)
						flag = True
		if not flag:
			current_merge.append(single_ver_or_range)
		merged = current_merge

	# Repeating until there are no more changes
	if len(sversions_and_sranges) != len(merged):
		return MergeVersions(merged)

	return merged
