#!/usr/bin/python3

#	ax.vul : Lib that helps to work with vulnerability IDs
#	Copyright (C) 2020-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from ax.ver import Y_REGEX

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

FIXES_REGEX  = re.compile(
	r'(?P<pre>.*)'
	r'(?P<label>fix(e[sd])?(:|for)?)'
	r'(?P<post>.*)?'
	)
NEGATION_REGEX = re.compile(r'(\snot?|n\'t)\s+$')
VUL_ID_REGEX = re.compile(r'\s*'
	fr'(?P<vul_id>(cve\-|bdu:){Y_REGEX}\-[0-9]{{4,7}})'
	r'(?P<delim>\s*((,\s?)?and|,|\s)\s*)?'
	)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Identify format of a given line

# Format 1 is a listing of vulnerability IDs placed right after a variation
# of a keyword ('fix', 'fixed' etc.)

# Format 2 is a multiline listing of vulnerability IDs, each line starts with
# the '+' or the '*' symbol followed by a listing of vulnerability IDs, the
# keyword is the same

def IdentifyFormat(current_fmt, line):

	if current_fmt == 0:

		match = re.match(FIXES_REGEX, line)

		if not match or not match.group('pre') or \
				re.search(NEGATION_REGEX, match.group('pre')) or \
				(match.group('post') and
					re.search(NEGATION_REGEX, match.group('post'))):
			return current_fmt, None

		if match.group('pre')[0] == '-' and \
				(not match.group('post') or match.group('post')[-1] == ':'):
			return 2, None

		line = ''
		for k in ('pre', 'post'):
			piece  = match.group(k)
			if piece:
				search = re.search(VUL_ID_REGEX, piece)
				if search:
					line += piece[search.start():]

		return 1, (line if line else None)

	if current_fmt == 2:

		search = re.search(r'^[\*\+]+', line)

		if not search:
			if re.search(VUL_ID_REGEX, line):
				return IdentifyFormat(0, line)
			return 2, None

		return current_fmt, line[len(search.group()):]

	return current_fmt, line

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get a set of vulnerability IDs from a given listing string

def GetIDs(listing):

	vul_ids = set()

	while True:

		if not listing:
			break

		match = re.match(VUL_ID_REGEX, listing)
		if not match:
			break

		vul_ids.add(match.group('vul_id'))
		listing = listing[len(match.group()):]

	return vul_ids, listing

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get IDs of vulnerabilities from a given multiline byte array

def GetVulIDs(byte_array):

	vul_ids = set()
	fmt = 0
	count = 0
	warn = ''

	# Iterating through all lines of a given byte array
	for line in byte_array.splitlines():

		count += 1

		if not line or line[0] == '*':
			fmt = 0
			continue

		try:
			line = line.decode('utf-8').strip().lower()
		except Exception as exc:
			warn += ('; ' if warn else '') + f'line {count}, {exc}'
			fmt = 0
			continue

		if not line or line[0] == '-':
			fmt = 0
			if not line:
				continue

		while True:

			fmt, line = IdentifyFormat(fmt, line)
			if fmt == 0 or line == None:
				break

			extracted_ids, line = GetIDs(line)
			vul_ids |= extracted_ids

			# Processing the leftovers
			if fmt == 1 and line:
				fmt = 0
				continue

			break

	return vul_ids, warn
